# distutils: language = c++
# cython: c_string_type=unicode, c_string_encoding=utf8

from gomea.real_valued cimport rvg_t, Config
from gomea.output cimport OutputStatisticsWrapper
from gomea.output import OutputStatistics
from libcpp.string cimport string
from libcpp cimport bool
import inspect

from gomea.fitness cimport FitnessFunction, GBOFitnessFunction
from gomea.linkage cimport LinkageModel, StaticLinkageTree
include "gomea/EmbeddedFitness.pxi"

# Create a Cython extension type which holds a C++ instance
# as an attribute and create a bunch of forwarding methods
# Python extension type.
cdef class RealValuedGOMEA:
    cdef rvg_t c_inst  # Hold a C++ instance which we're wrapping
    cdef Config c_config

    def __cinit__(self,
        # Optimization problem settings (required)
        fitness: FitnessFunction, 
        # GOMEA parameters
        linkage_model : LinkageModel = StaticLinkageTree(),
        lower_init_range: double=0.0,
        upper_init_range: double=1.0,
        random_seed: int=-1,
        # IMS settings (optional)
        max_number_of_populations: int=25,
        max_number_of_generations: int=-1,
        IMS_subgeneration_factor: int=8,
        base_population_size: int=10,
        # Termination settings (optional)
        max_number_of_evaluations : double=-1,
        max_number_of_seconds : double=-1,
    ):
        
        # Initialize attributes 
        self.c_config = Config()
        self.c_config.problem_index = 0
        assert( (<FitnessFunction?>fitness).c_inst_realvalued != NULL, "FitnessFunction is not real-valued." )
        self.c_config.fitness = (<FitnessFunction?>fitness).c_inst_realvalued
        self.c_config.linkage_config = linkage_model.c_inst
        self.c_config.use_vtr = True 
        self.c_config.lower_user_range = lower_init_range
        self.c_config.upper_user_range = upper_init_range
        self.c_config.tau = 0.35
        self.c_config.distribution_multiplier_decrease = 0.9
        self.c_config.st_dev_ratio_threshold = 1.0
        self.c_config.fitness_variance_tolerance = 0.0
        self.c_config.maximum_no_improvement_stretch = 100
        self.c_config.selection_during_gom = True
        self.c_config.update_elitist_during_gom = True 
        self.c_config.maximum_number_of_populations = max_number_of_populations
        self.c_config.number_of_subgenerations_per_population_factor = IMS_subgeneration_factor
        self.c_config.base_population_size = base_population_size
        self.c_config.maximum_number_of_generations = max_number_of_generations
        self.c_config.maximum_number_of_evaluations = max_number_of_evaluations
        self.c_config.maximum_number_of_seconds = max_number_of_seconds
        self.c_config.black_box_evaluations = False
        self.c_config.write_generational_statistics = True 
        self.c_config.write_generational_solutions = False
        self.c_config.verbose = False
        self.c_config.print_verbose_overview = False 
        self.c_config.fix_seed = 0
        if random_seed != -1:
            self.c_config.random_seed = random_seed
            self.c_config.fix_seed = 1

        # Initialize C++ instance
        self.c_inst = rvg_t(&self.c_config)
    
    def print_usage(self):
        self.c_inst.printUsage()

    def run(self):
        self.c_inst.run()
        return OutputStatistics(OutputStatisticsWrapper.from_ptr(&self.c_inst.output))



from libcpp.string cimport string
from libcpp cimport bool
from gomea.output cimport output_statistics_t
from gomea.fitness cimport fitness_t
from gomea.linkage cimport linkage_config_t 

cdef extern from "gomea/src/discrete/Config.hpp" namespace "gomea::discrete":
    cdef cppclass Config:
        Config() except +

        int maximumNumberOfGOMEAs, IMSsubgenerationFactor, basePopulationSize, maxArchiveSize, maximumNumberOfEvaluations, maximumNumberOfGenerations, AnalyzeFOS
        long long randomSeed
        bool fix_seed
        double maximumNumberOfSeconds
        string folder, problemInstancePath
        fitness_t[char] *fitness
        linkage_config_t *linkage_config

cdef extern from "gomea/src/discrete/gomeaIMS.hpp" namespace "gomea::discrete":
    cdef cppclass gomeaIMS:
        gomeaIMS() except +
        gomeaIMS(Config*) except +
        
        void run() except +
        void runGeneration() except +
        bool checkTermination() except +
        double getProgressUntilTermination() except +
        output_statistics_t output



from libcpp.string cimport string
from libcpp cimport bool
from gomea.output cimport output_statistics_t
from gomea.fitness cimport fitness_t
from gomea.linkage cimport linkage_config_t

cdef extern from "src/real_valued/Config.hpp" namespace "gomea::realvalued":
    cdef cppclass Config:
        Config() except +

        int problem_index, maximum_no_improvement_stretch, maximum_number_of_populations, number_of_subgenerations_per_population_factor, base_population_size, maximum_number_of_generations 
        double lower_user_range, upper_user_range, tau, distribution_multiplier_decrease, st_dev_ratio_threshold, fitness_variance_tolerance, maximum_number_of_evaluations, maximum_number_of_seconds 
        bool fix_seed, use_vtr, write_generational_statistics, write_generational_solutions, black_box_evaluations, selection_during_gom, update_elitist_during_gom, verbose, print_verbose_overview
        long long random_seed
        linkage_config_t *linkage_config
        fitness_t[double] *fitness

cdef extern from "src/real_valued/rv-gomea.hpp" namespace "gomea::realvalued":
    cdef cppclass rvg_t:
        rvg_t() except +
        rvg_t(Config*) except +
        
        void printUsage() except +
        void run() except +
        
        output_statistics_t output



# distutils: language = c++

import ctypes
import numpy as np
cimport numpy as np
from libcpp.vector cimport vector

np.import_array()

cdef public double gomea_pyfitness_subfunction_realvalued(obj, int subfunction_index, vector[double] &variables ) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_FLOAT64, vec_ptr)
    
    cdef double subf = fitness_obj.subfunction(subfunction_index,npvars)
    return subf

cdef public double gomea_pyfitness_subfunction_discrete(obj, int subfunction_index, vector[char] &variables ) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_BYTE, vec_ptr)
    
    cdef double subf = fitness_obj.subfunction(subfunction_index,npvars)
    return subf

cdef public double gomea_pyfitness_objective_function_gbo(obj, int objective_index, vector[double] &fitness_buffers ) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    
    cdef void *vec_ptr = &fitness_buffers[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> fitness_buffers.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_FLOAT64, vec_ptr)
    
    cdef double result = fitness_obj.objective_function(objective_index,npvars)
    return result

cdef public double gomea_pyfitness_constraint_function_gbo(obj, vector[double] &fitness_buffers ) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    
    cdef void *vec_ptr = &fitness_buffers[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> fitness_buffers.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_FLOAT64, vec_ptr)
    
    cdef double result = fitness_obj.constraint_function(npvars)
    return result 

cdef public double gomea_pyfitness_objective_function_bbo_discrete(obj, int objective_index, vector[char] &variables ) except +:
    fitness_obj = <FitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_BYTE, vec_ptr)
    
    cdef double result = fitness_obj.objective_function(objective_index,npvars)
    return result

cdef public double gomea_pyfitness_objective_function_bbo_realvalued(obj, int objective_index, vector[double] &variables ) except +:
    fitness_obj = <FitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_FLOAT64, vec_ptr)
    
    cdef double result = fitness_obj.objective_function(objective_index,npvars)
    return result

cdef public double gomea_pyfitness_constraint_function_bbo_discrete(obj, vector[char] &variables ) except +:
    fitness_obj = <FitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_BYTE, vec_ptr)
    
    cdef double result = fitness_obj.constraint_function(npvars)
    return result 

cdef public double gomea_pyfitness_constraint_function_bbo_realvalued(obj, vector[double] &variables ) except +:
    fitness_obj = <FitnessFunction?>obj
    
    cdef void *vec_ptr = &variables[0]
    cdef np.npy_intp shape[1]
    shape[0] = <np.npy_intp> variables.size()
    cdef np.ndarray npvars = np.PyArray_SimpleNewFromData(1, shape, np.NPY_FLOAT64, vec_ptr)
    
    cdef double result = fitness_obj.constraint_function(npvars)
    return result 

cdef public vector[int] gomea_pyfitness_inputsToSubfunction(obj, int subfunction_index ) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    cdef vector[int] indices = fitness_obj.inputs_to_subfunction(subfunction_index)
    return indices 

cdef public int gomea_pyfitness_index_of_fitness_buffer(obj, int subfunction_index) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    cdef int result = fitness_obj.fitness_buffer_index_for_subfunction(subfunction_index)
    return result

cdef public int gomea_pyfitness_number_of_fitness_buffers(obj) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    cdef int n = fitness_obj.number_of_fitness_buffers()
    return n

cdef public int gomea_pyfitness_numberOfSubfunctions(obj) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    cdef int n = fitness_obj.number_of_subfunctions()
    return n

cdef public double gomea_pyfitness_similarity_measure(obj, size_t var_a, size_t var_b) except +:
    fitness_obj = <GBOFitnessFunction?>obj
    cdef double result = fitness_obj.similarity_measure(var_a,var_b)
    return result



VERSION_MAJOR = 1
VERSION_MINOR = 0
VERSION_PATCH = 2
VERSION_STRING = '%s.%s.%s' % (VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH)
__version__ = VERSION_STRING

from setuptools import Extension, setup, find_namespace_packages
from Cython.Build import cythonize
import sys
import glob
import numpy as np
import platform

if sys.version_info[0] == 2:
        raise Exception('Python 2.x is no longer supported')

with open("README.md", 'r') as f:
        long_description = f.read()

debug_mode = False
if '--debug' in sys.argv:
        debug_mode = True

common_src = glob.glob("gomea/src/common/*.cpp") + glob.glob("gomea/src/utils/*.cpp")
fitness_src = glob.glob("gomea/src/fitness/*.cpp") + glob.glob("gomea/src/fitness/benchmarks-rv/*.cpp") + glob.glob("gomea/src/fitness/benchmarks-discrete/*.cpp")

compile_args = ["-std=c++17"]
link_args = ["-std=c++17"]
if debug_mode:
        compile_args.extend(['-UNDEBUG','-g'])
        link_args.extend(['-UNDEBUG','-g'])
else:
        compile_args.extend(['-O3','-g0'])
        link_args.extend(['-O3','-g0'])
if platform.system() == "Darwin":
        compile_args.extend(["-stdlib=libc++","-mmacosx-version-min=10.15"])
        link_args.extend(["-stdlib=libc++","-mmacosx-version-min=10.15"])

if platform.system() == "Windows":
        compile_args = ["/std:c++17"]
        link_args = []
        if debug_mode:
                compile_args.extend(['/UNDEBUG','/Zi','/g0'])
        else:
                compile_args.extend(['/O2'])

extensions = []

extensions.append( Extension("gomea.discrete",
        ["gomea/discrete.pyx"] + glob.glob("gomea/src/discrete/*.cpp") + common_src + fitness_src,
        include_dirs = ["."] + ["gomea/lib/cxxopts-3.1.1/include/"] + [np.get_include()],
        language="c++",
        extra_compile_args=compile_args,
        extra_link_args=link_args)
)

extensions.append( Extension("gomea.real_valued",
        ["gomea/real_valued.pyx"] + glob.glob("gomea/src/real_valued/*.cpp") + common_src + fitness_src,
        include_dirs = ["."] + ["gomea/lib/Eigen"] + [np.get_include()],
        language="c++",
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        library_dirs=[],
        extra_objects=[])
)

extensions.append( Extension("gomea.fitness",
        ["gomea/fitness.pyx"] + fitness_src + common_src,
        include_dirs = ["."] + [np.get_include()],
        language="c++",
        extra_compile_args=compile_args,
        extra_link_args=link_args)
)

extensions.append( Extension("gomea.linkage",
        ["gomea/linkage.pyx"] + common_src,
        include_dirs = ["."] + [np.get_include()],
        language="c++",
        extra_compile_args=compile_args,
        extra_link_args=link_args)
)

extensions.append( Extension("gomea.output",
        ["gomea/output.pyx"] + common_src,
        include_dirs = ["."] + [np.get_include()],
        language="c++",
        extra_compile_args=compile_args,
        extra_link_args=link_args)
)

setup(
    name = "gomea",
    description = 'Library for the use of various variants of the Gene-pool Optimal Mixing Evolutionary Algorith (GOMEA).',
    author = 'Anton Bouter',
    author_email = 'Anton.Bouter@cwi.nl',
    url = 'https://github.com/abouter/gomea',
    version = __version__,
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    packages=["gomea"],
    include_package_data=True,
    include_dirs=["gomea"],
    ext_modules = cythonize(extensions,
        include_path = ["."] + [np.get_include()],
        gdb_debug = debug_mode,
        language_level = "3"),
    install_requires=["numpy>=1.23.0","tqdm"],
    zip_safe = False
)

